import pytest

from broqer import op, NONE, Value
from tests.helper_single import check_get_method, check_subscription, check_dependencies


def add(a, b, c, constant=0):
    return a + b * 2 + c * 3 + constant * 4


test_vector = [
    # o, args, kwargs, input_vector, output_vector
    (op.Map, (lambda v: v+1,), {}, [NONE, 1, 2], [NONE, 2, 3]),
    (op.Map, (lambda v: NONE,), {}, [1, 2], [NONE, NONE]),
    (op.Map, (), {'function': lambda v: v+1}, [NONE, 1, 2], [NONE, 2, 3]),
    (op.Map, (lambda a, b: a+b,), {'unpack': True}, [NONE, (1, 1), (2, 2)], [NONE, 2, 4]),
    (op.Map, (add, 1), {'unpack': True, 'constant': 2}, [NONE, (1, 1), (2, 2)], [NONE, 14, 19]),
    (op.build_map_factory(lambda v, o: v+o,), (1,), {}, [NONE, 1, 2], [NONE, 2, 3]),
    (op.build_map_factory(lambda a, b: a+b, unpack=True), (), {}, [(1, 1), (2, 2)], [2, 4]),
    (op.build_map_factory(unpack=True)(lambda a, b: a+b), (), {}, [(1, 1), (2, 2)], [2, 4]),
    (lambda: op.build_map(lambda v: v+1,), (), {}, [NONE, 1, 2], [NONE, 2, 3]),
    (lambda: op.build_map(unpack=False)(lambda v: v+1,), (), {}, [NONE, 1, 2], [NONE, 2, 3]),
]

@pytest.mark.parametrize('method', [check_get_method, check_subscription, check_dependencies])
@pytest.mark.parametrize('o,args,kwargs,input_vector,output_vector', test_vector)
def test_operator(method, o, args, kwargs, input_vector, output_vector):
    operator = o(*args, **kwargs)

    method(operator, input_vector, output_vector)


def test_map_factory_keyword():
    m = op.build_map_factory(lambda v: v+1)
    v = Value()
    with pytest.raises(TypeError, message='"unpack" has to be defined by decorator'):
        o = v | m(unpack=True)


""" Module containing Subscribers """
from .on_emit_future import OnEmitFuture
from .sink import Sink, build_sink, build_sink_factory, sink_property
from .sink_async import SinkAsync, build_sink_async, \
                        build_sink_async_factory, sink_async_property
from .trace import Trace

__all__ = ['OnEmitFuture', 'Sink', 'build_sink',
           'build_sink_factory', 'sink_property', 'SinkAsync',
           'build_sink_async', 'build_sink_async_factory',
           'sink_async_property', 'Trace']



import asyncio
import pytest
from unittest import mock

from broqer.publishers import PollPublisher
from broqer import Sink, NONE

from .eventloop import VirtualTimeEventLoop


@pytest.yield_fixture()
def event_loop():
    loop = VirtualTimeEventLoop()
    yield loop
    loop.close()


@pytest.mark.asyncio
async def test_subscribe(event_loop):
    poll_mock = mock.Mock(return_value=3)
    sink_mock = mock.Mock()

    p = PollPublisher(poll_mock, 1)

    await asyncio.sleep(1, loop=event_loop)
    assert p.get() is NONE

    p.subscribe(Sink(sink_mock))
    sink_mock.assert_called_once_with(3)
    poll_mock.assert_called_once()

    await asyncio.sleep(2.5, loop=event_loop)
    sink_mock.assert_called_with(3)
    assert sink_mock.call_count == 3


""" Implementation of abstract Disposable.
"""
from abc import ABCMeta, abstractmethod


class Disposable(metaclass=ABCMeta):
    """
    Implementation of the disposable pattern. A disposable is usually
    returned on resource allocation. Calling .dispose() on the returned
    disposable is freeing the resource.

    Note: Multiple calls to .dispose() have to be handled by the
    implementation.

        >>> class MyDisposable(Disposable):
        ...     def dispose(self):
        ...         print('DISPOSED')

        >>> with MyDisposable():
        ...     print('working')
        working
        DISPOSED
    """
    @abstractmethod
    def dispose(self) -> None:
        """ .dispose() method has to be overwritten"""

    def __enter__(self):
        """ Called on entry of a new context """
        return self

    def __exit__(self, _type, _value, _traceback):
        """ Called on exit of the context. .dispose() is called here """
        self.dispose()



"""
Rate limit emits by the given time.
Usage:
>>> import asyncio
>>> from broqer import Value, op, Sink
>>> v = Value()
>>> throttle_publisher = v | op.Throttle(0.1)
>>> _d = throttle_publisher.subscribe(Sink(print))
>>> v.emit(1)
1
>>> v.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
>>> v.emit(3)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.2))
3
>>> # It's also possible to reset the throttling duration:
>>> v.emit(4)
4
>>> v.emit(5)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.05))
>>> throttle_publisher.reset()
"""
import asyncio
import sys
from typing import Any  # noqa: F401

from broqer import Publisher, default_error_handler, NONE

from broqer.operator import Operator, OperatorFactory
from broqer.timer import Timer


class AppliedThrottle(Operator):
    """ Rate limit emits by the given time.
    :param duration: time for throttling in seconds
    :param error_callback: the error callback to be registered
    :param loop: asyncio event loop to use
    """
    def __init__(self, publisher: Publisher, duration: float,
                 error_callback=default_error_handler, loop=None) -> None:

        Operator.__init__(self, publisher)

        self._duration = duration
        self._loop = loop or asyncio.get_event_loop()
        self._timer = Timer(self._delayed_emit_cb, loop=loop)
        self._error_callback = error_callback

    def get(self):
        return Publisher.get(self)

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if not self._timer.is_running():
            self._timer.start(timeout=0, args=(value,))
        else:
            self._timer.change_arguments(args=(value,))

    def _delayed_emit_cb(self, value=NONE):
        if value is NONE:
            # since the last emit the given duration has passed without another
            # emit
            return

        try:
            Publisher.notify(self, value)
        except Exception:  # pylint: disable=broad-except
            self._error_callback(*sys.exc_info())

        self._timer.start(self._duration)

    def reset(self):
        """ Reseting duration for throttling """
        self._timer.cancel()


class Throttle(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Apply throttling to each emitted value.
    :param duration: time for throttling in seconds
    """
    def __init__(self, duration: float) -> None:
        if duration < 0:
            raise ValueError('Duration has to be bigger than zero')

        self._duration = duration

    def apply(self, publisher: Publisher):
        return AppliedThrottle(publisher, self._duration)



""" This module enables the operator overloading of publishers """
import math
import operator
from typing import Any as Any_

# pylint: disable=cyclic-import
import broqer
from broqer import Publisher
from broqer.operator import Operator


class MapConstant(Operator):
    """ MapConstant TODO Docstring """
    def __init__(self, publisher: Publisher, value, operation) -> None:
        Operator.__init__(self, publisher)
        self._value = value
        self._operation = operation

        if publisher.inherited_type is not None:
            self.inherit_type(publisher.inherited_type)

    def get(self):
        return self._operation(self._orginator.get(), self._value)

    def emit(self, value: Any_, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        result = self._operation(value, self._value)

        return Publisher.notify(self, result)


class MapConstantReverse(Operator):
    """ MapConstantReverse TODO """
    def __init__(self, publisher: Publisher, value, operation) -> None:
        Operator.__init__(self, publisher)
        self._value = value
        self._operation = operation

        if publisher.inherited_type is not None:
            self.inherit_type(publisher.inherited_type)

    def get(self):
        return self._operation(self._value, self._orginator.get())

    def emit(self, value: Any_, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        result = self._operation(self._value, value)

        return Publisher.notify(self, result)


class MapUnary(Operator):
    """ MapUnary TODO """
    def __init__(self, publisher: Publisher, operation) -> None:
        Operator.__init__(self, publisher)
        self._operation = operation

        if publisher.inherited_type is not None:
            self.inherit_type(publisher.inherited_type)

    def get(self):
        return self._operation(self._orginator.get())

    def emit(self, value: Any_, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        result = self._operation(value)

        return Publisher.notify(self, result)


class _GetAttr(Operator):
    def __init__(self, publisher: Publisher, attribute_name) -> None:
        Operator.__init__(self, publisher)
        self._attribute_name = attribute_name
        self._args = None
        self._kwargs = None

        self.inherit_type(publisher.inherited_type)

    def get(self):
        value = self._orginator.get()  # may raise ValueError
        attribute = getattr(value, self._attribute_name)
        if self._args is None:
            return attribute
        return attribute(*self._args, **self._kwargs)

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        return self

    def emit(self, value: Any_, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        attribute = getattr(value, self._attribute_name)

        if self._args is None:
            return Publisher.notify(self, attribute)

        return Publisher.notify(self, attribute(*self._args, **self._kwargs))


def apply_operator_overloading():
    """ Function to apply operator overloading to Publisher class """
    # operator overloading is (unfortunately) not working for the following
    # cases:
    # int, float, str - should return appropriate type instead of a Publisher
    # len - should return an integer
    # 'x in y' - is using __bool__ which is not working with Publisher
    for method in (
            '__lt__', '__le__', '__eq__', '__ne__', '__ge__', '__gt__',
            '__add__', '__and__', '__lshift__', '__mod__', '__mul__',
            '__pow__', '__rshift__', '__sub__', '__xor__', '__concat__',
            '__getitem__', '__floordiv__', '__truediv__'):
        def _op(operand_left, operand_right, operation=method):
            if isinstance(operand_right, Publisher):
                return broqer.op.CombineLatest(operand_left, operand_right,
                                               map_=getattr(operator,
                                                            operation))
            return MapConstant(operand_left, operand_right,
                               getattr(operator, operation))

        setattr(Publisher, method, _op)

    for method, _method in (
            ('__radd__', '__add__'), ('__rand__', '__and__'),
            ('__rlshift__', '__lshift__'), ('__rmod__', '__mod__'),
            ('__rmul__', '__mul__'), ('__rpow__', '__pow__'),
            ('__rrshift__', '__rshift__'), ('__rsub__', '__sub__'),
            ('__rxor__', '__xor__'), ('__rfloordiv__', '__floordiv__'),
            ('__rtruediv__', '__truediv__')):
        def _op(operand_left, operand_right, operation=_method):
            return MapConstantReverse(operand_left, operand_right,
                                      getattr(operator, operation))

        setattr(Publisher, method, _op)

    for method, _method in (
            ('__neg__', operator.neg), ('__pos__', operator.pos),
            ('__abs__', operator.abs), ('__invert__', operator.invert),
            ('__round__', round), ('__trunc__', math.trunc),
            ('__floor__', math.floor), ('__ceil__', math.ceil)):
        def _op_unary(operand, operation=_method):
            return MapUnary(operand, operation)

        setattr(Publisher, method, _op_unary)

    def _getattr(publisher, attribute_name):
        if not publisher.inherited_type or \
           not hasattr(publisher.inherited_type, attribute_name):
            raise AttributeError('Attribute %r not found' % attribute_name)
        return _GetAttr(publisher, attribute_name)

    setattr(Publisher, '__getattr__', _getattr)



from unittest import mock
import pytest

from broqer import Disposable, Publisher, Value, Sink, Trace, build_sink, \
                   build_sink_factory


@pytest.mark.parametrize('operator_cls', [Sink, Trace])
def test_sink(operator_cls):
    cb = mock.Mock()

    s = Value()
    sink_instance = s.subscribe(operator_cls(cb))
    assert isinstance(sink_instance, Disposable)

    assert not cb.called
    assert len(s.subscriptions) == 1

    # test various emits on source
    with pytest.raises(TypeError):
        s.emit()

    s.emit(None)
    cb.assert_called_with(None)

    s.emit(1)
    cb.assert_called_with(1)

    s.emit((1, 2))
    cb.assert_called_with((1, 2))

    # testing dispose()s
    cb.reset_mock()

    sink_instance.dispose()
    assert len(s.subscriptions) == 0

    s.emit(1)
    assert not cb.called


@pytest.mark.parametrize('operator_cls', [Sink, Trace])
def test_sink2(operator_cls):
    cb = mock.Mock()

    s = Value()
    sink_instance = s.subscribe(operator_cls(cb, unpack=True))
    assert isinstance(sink_instance, Disposable)

    # test various emits on source
    with pytest.raises(TypeError):
        s.emit()

    with pytest.raises(TypeError):
        s.emit(1)

    cb.assert_not_called()

    s.emit((1, 2))
    cb.assert_called_with(1, 2)


@pytest.mark.parametrize('operator_cls', [Sink, Trace])
def test_sink_without_function(operator_cls):
    s = Value()
    sink_instance = s.subscribe(operator_cls())
    assert isinstance(sink_instance, Disposable)
    assert len(s.subscriptions) == 1

    s.emit(1)


@pytest.mark.parametrize('operator', [Sink, Trace])
def test_sink_on_subscription(operator):
    cb = mock.Mock()

    s = Value(0)
    sink_instance = s.subscribe(operator(cb))
    assert isinstance(sink_instance, Disposable)

    cb.assert_called_with(0)
    assert len(s.subscriptions) == 1

    s.emit(1)
    cb.assert_called_with(1)

    # testing dispose()
    cb.reset_mock()

    sink_instance.dispose()
    assert len(s.subscriptions) == 0

    s.emit(1)
    assert not cb.called


@pytest.mark.parametrize('operator_cls', [Sink, Trace])
def test_sink_partial(operator_cls):
    cb = mock.Mock()

    s = Value()
    sink_instance = s.subscribe(operator_cls(cb, 1, 2, 3, a=1))
    assert isinstance(sink_instance, Disposable)

    assert not cb.called
    assert len(s.subscriptions) == 1

    # test various emits on source
    s.emit(None)
    cb.assert_called_with(1, 2, 3, None, a=1)

    s.emit(1)
    cb.assert_called_with(1, 2, 3, 1, a=1)

    s.emit((1, 2))
    cb.assert_called_with(1, 2, 3, (1, 2), a=1)

    # testing dispose()
    cb.reset_mock()

    sink_instance.dispose()
    assert len(s.subscriptions) == 0

    s.emit(1)
    assert not cb.called


@pytest.mark.parametrize(
    'build_kwargs, init_args, init_kwargs, ref_args, ref_kwargs, exception',
    [(None, (), {}, (), {}, None),
     ({'unpack': True}, (), {}, (), {'unpack': True}, None),
     ({'unpack': False}, (), {}, (), {'unpack': False}, None),
     ({'unpack': False}, (), {'unpack': False}, (), {'unpack': False},
      TypeError),
     (None, (1,), {'a': 2}, (1,), {'unpack': False, 'a': 2}, None),
     ({'unpack': True}, (1,), {'a': 2}, (1,), {'unpack': True, 'a': 2}, None),
     ({'unpack': False}, (1,), {'a': 2}, (1,), {'unpack': False, 'a': 2},
      None),
     ({'foo': 1}, (), {}, (), {}, TypeError),
     ])
def test_build(build_kwargs, init_args, init_kwargs, ref_args, ref_kwargs,
               exception):
    mock_cb = mock.Mock()
    ref_mock_cb = mock.Mock()

    reference = Sink(ref_mock_cb, *ref_args, **ref_kwargs)

    try:
        if build_kwargs is None:
            dut = build_sink_factory(mock_cb)(*init_args, **init_kwargs)
        else:
            dut = build_sink_factory(**build_kwargs)(mock_cb)(*init_args,
                                                      **init_kwargs)
    except Exception as e:
        assert isinstance(e, exception)
        return
    else:
        assert exception is None

    assert dut._unpack == reference._unpack

    v = Publisher((1, 2))
    v.subscribe(dut)
    v.subscribe(reference)

    assert mock_cb.mock_calls == ref_mock_cb.mock_calls
    assert len(mock_cb.mock_calls) == 1



"""
Cache the latest emit - the result is suppressing multiple emits with the same
value. Also initialization can be defined in the case the source publisher does
not emit on subscription.

Usage:

>>> from broqer import Value, op, Sink
>>> s = Value(1)

>>> cached_publisher = s | op.Cache()
>>> _disposable = cached_publisher.subscribe(Sink(print))
1
>>> s.emit(2)
2
>>> s.emit(2)
>>> _disposable.dispose()

Using the initial value for cache:

>>> from broqer import Value, op, Sink
>>> s = Value()

>>> cached_publisher = s | op.Cache(1)
>>> _disposable = cached_publisher.subscribe(Sink(print))
1
>>> s.emit(1)
>>> s.emit(2)
2
>>> _disposable.dispose()
"""
from typing import Any

from broqer import Publisher, NONE
from broqer.publisher import TValue
from broqer.operator import Operator, OperatorFactory


class AppliedCache(Operator):
    """ Cache object applied to publisher (see Map) """
    def __init__(self, publisher: Publisher, init: Any = NONE) -> None:
        Operator.__init__(self, publisher)
        self._state = init

    def get(self) -> TValue:
        return self._orginator.get()

    def emit(self, value: TValue, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if value != self._state:
            self._state = value
            return Publisher.notify(self, value)

        return None


class Cache(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Cache emitted values. Suppress duplicated value emits.

    :param init: optional initialization state
    """
    def __init__(self, init: Any = NONE) -> None:
        self._init = init

    def apply(self, publisher: Publisher):
        return AppliedCache(publisher, self._init)



import sys
from unittest.mock import Mock

import pytest

from broqer.error_handler import default_error_handler


def test_default(capsys):
    try:
        0/0
    except:
        exc = sys.exc_info()

    default_error_handler(*exc)

    captured = capsys.readouterr()

    assert captured.err.startswith('Traceback (most recent call last):')
    assert 'ZeroDivisionError: division by zero' in captured.err
    assert __file__ in captured.err

    assert captured.out == ''


def test_set_errorhandler(capsys):
    mock = Mock()
    default_error_handler.set(mock)

    try:
        0/0
    except:
        exc = sys.exc_info()

    default_error_handler(*exc)

    mock.assert_called_once_with(*exc)

    captured = capsys.readouterr()

    assert captured.err == ''
    assert captured.out == ''

    # reset
    default_error_handler.reset()

    default_error_handler(*exc)

    captured = capsys.readouterr()
    assert captured.err.startswith('Traceback (most recent call last):')



""" Asynchronous timer object """
import asyncio
from typing import Callable, Optional


class Timer:
    """ The timer object is used to have an abstraction for handling time
    dependend functionality. The timer works non-periodically.

    :param callback: an optional callback function called when the time
                     `timeout` has passed after calling `.start()` or when
                     calling `.end_early()`
    :param loop: optional asyncio event loop
    """
    def __init__(self, callback: Optional[Callable[[], None]] = None,
                 loop: Optional[asyncio.BaseEventLoop] = None):
        self._callback = callback
        self._handle = None  # type: Optional[asyncio.Handle]
        self._loop = loop or asyncio.get_event_loop()
        self._args = None

    def start(self, timeout: float, args=()) -> None:
        """ start the timer with given timeout. Optional arguments for the
        callback can be provided. When the timer is currently running, the
        timer will be re-set to the new timeout.

        :param timeout: time in seconds to the end of the timer
        :param args: optional tuple with arguments for the callback
        """
        if self._handle:
            self._handle.cancel()

        self._args = args

        if timeout > 0:
            self._handle = self._loop.call_later(timeout, self._trigger)
        else:
            self._trigger()

    def change_arguments(self, args=()):
        """ Will chance the arguments for the scheduled callback

        :param args: Positional arguments
        """
        self._args = args

    def cancel(self) -> None:
        """ Cancel the timer. An optional callback will not be called. """

        if self._handle:
            self._handle.cancel()
        self._handle = None

    def end_early(self) -> None:
        """ immediate stopping the timer and call optional callback """
        self._handle = None
        if self._handle and self._callback:
            self._callback(*self._args)

    def is_running(self) -> bool:
        """ tells if the timer is currently running
        :returns: boolean, True when timer is running
        """
        return self._handle is not None

    def _trigger(self):
        """ internal method called when timer is finished """
        self._handle = None
        if self._callback:
            self._callback(*self._args)



""" Implementing Value """

from typing import Any

# pylint: disable=cyclic-import
from broqer import Publisher, Subscriber, NONE


class Value(Publisher, Subscriber):
    """
    Value is a publisher and subscriber.

    >>> from broqer import Sink

    >>> s = Value(0)
    >>> _d = s.subscribe(Sink(print))
    0
    >>> s.emit(1)
    1
    """
    def __init__(self, init=NONE):
        Publisher.__init__(self, init)
        Subscriber.__init__(self)

    def emit(self, value: Any,
             who: Publisher = None) -> None:  # pylint: disable=unused-argument
        return Publisher.notify(self, value)



from unittest import mock

import pytest

from broqer import Publisher, NONE, op, Sink


def check_get_method(operator, input_vector, output_vector):
    input_value, output_value = input_vector[0], output_vector[0]

    p = Publisher()
    p.get = mock.MagicMock(return_value=input_value)
    o = p | operator

    with pytest.raises(ValueError):
        o.emit(input_value, who=None)

    assert o.get() == output_value

    disposable = o.subscribe(Sink())

    if input_value is not NONE:
        o.emit(input_value, who=p)  # simulate emit on subscribe

    p.get = None  # .get should not be called when operator has a subscription

    assert o.get() == output_value  # this should retrieve the internal state

    # after unsubscription it should remove the internal state and retrieve it
    # directly from orignator publisher via get
    disposable.dispose()

    p.get = mock.MagicMock(return_value=input_value)

    assert o.get() == output_value

    p.get.assert_called_once_with()


def check_subscription(operator, input_vector, output_vector):
    assert len(input_vector) == len(output_vector)

    m = mock.Mock()

    p = Publisher(input_vector[0])
    o = p | operator

    o2 = Publisher() | operator

    # subscribe operator to publisher
    disposable = o.subscribe(Sink(m))

    assert p.subscriptions == (o,)  # now it should be subscribed
    assert o.dependencies == (p,)

    # test emit on subscription
    if output_vector[0] is not NONE:
        m.assert_called_once_with(output_vector[0])

    m.reset_mock()

    # test input_vector
    for input_value, output_value in zip(input_vector[1:], output_vector[1:]):
        if input_value is not NONE:
            p.notify(input_value)

        if output_value is NONE:
            m.assert_not_called()
        else:
            m.assert_called_with(output_value)
        m.reset_mock()

    # test input_vector with unsubscriptions between
    disposable.dispose()
    assert o.dependencies == (p,)
    disposable = o.subscribe(Sink(m))

    for input_value, output_value in zip(input_vector, output_vector):
        if input_value is not NONE:
            m.reset_mock()
            p.notify(input_value)
            if output_value is NONE:
                m.assert_not_called()
            else:
                m.assert_called_once_with(output_value)


def check_dependencies(operator, *_):
    p = Publisher(NONE)

    assert len(p.subscriptions) == 0

    o = p | operator

    assert p.subscriptions == ()  # operator should not be subscribed yet
    assert o.dependencies == (p,)
    assert o.subscriptions == ()

    # subscribe to operator
    disposable1 = o.subscribe(Sink())

    assert p.subscriptions == (o,)  # operator should now be subscriped
    assert o.dependencies == (p,)
    assert o.subscriptions == (disposable1.subscriber,)

    # second subscribe to operator
    disposable2 = o.subscribe(Sink(), prepend=True)

    assert p.subscriptions == (o,)
    assert o.dependencies == (p,)
    assert o.subscriptions == (disposable2.subscriber, disposable1.subscriber)

    # remove first subscriber
    disposable1.dispose()

    assert p.subscriptions == (o,)
    assert o.dependencies == (p,)
    assert o.subscriptions == (disposable2.subscriber,)

    # remove second subscriber
    disposable2.dispose()

    assert p.subscriptions == ()  # operator should now be subscriped
    assert o.dependencies == (p,)
    assert o.subscriptions == ()

